# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen
"""Class to represent files."""

from __future__ import annotations
from datetime import datetime, timezone
import hashlib
import os
from pathlib import Path, PurePosixPath
from typing import Optional, Union

import dateutil.parser
from pyscicat.model import DataFile

from .error import IntegrityError
from .logging import get_logger
from .typing import Downloader


class File:
    """Store local and remote paths and metadata for a file.

    There are two central properties:

    - ``remote_access_path``: Full path to the remote file if the file exists
      on the file server. Is ``None`` if the file does not exist on the remote.
    - ``local_path``: Path to the file on the local filesystem if it exists.
      Is ``None`` if the file does not exist locally.
    """

    def __init__(
        self,
        *,
        source_folder: Optional[str],
        local_path: Optional[Union[str, Path]],
        model: DataFile,
    ):
        self._local_path = Path(local_path) if local_path is not None else None
        self._source_folder = str(source_folder) if source_folder is not None else None
        # remote_path is stored as model.path
        self._model = model

    @classmethod
    def from_local(
        cls,
        path: Union[str, Path],
        *,
        base_path: Union[str, Path] = "",
        remote_path: Optional[Union[str, PurePosixPath]] = None,
        source_folder: Optional[Union[str, PurePosixPath]] = None,
        checksum_algorithm: str = "md5",
        remote_uid: Optional[str] = None,
        remote_gid: Optional[str] = None,
        remote_perm: Optional[str] = None,
    ) -> File:
        """Link a file on the local filesystem.

        Given following ``path``, ``base_path``, and ``source_folder`` ::

            path:                      somewhere/on/local/folder/file.nxs
            base_bath:                 somewhere/on/local
            source_folder:             remote/storage

        the file will be located on the remote at ::

            -> remote_path:            folder/file.nxs
            -> actual remote location: remote/storage/folder/file.nxs

        ``remote_path`` can also be overriden, in which case ``path`` and
        ``base_path`` are not used to deduce the "actual remote location".

        Parameters
        ----------
        path:
            Full path the local file.
        base_path:
            Only use ``path.relative_to(base_path)`` to determine the remote location.
        remote_path:
            Path on the remote, relative to ``source_folder``.
        source_folder:
            Base path on the remote.
            Must be ``None`` if the file does not exist on the remote.
        checksum_algorithm:
            Algorithm used to compute the file's checksum. See :mod:`hashlib`.
        remote_uid:
            User ID on the remote. Will be determined automatically on upload.
        remote_gid:
            Group ID on the remote. Will be determined automatically on upload.
        remote_perm:
            File permissions on the remote. Will be determined automatically on upload.
        """
        path = Path(path)
        if not remote_path:
            remote_path = path.relative_to(base_path)
        return File(
            local_path=path,
            source_folder=source_folder,
            model=file_model_from_local_file(
                path,
                remote_path=remote_path,
                checksum_algorithm=checksum_algorithm,
                uid=remote_uid,
                gid=remote_gid,
                perm=remote_perm,
            ),
        )

    @classmethod
    def from_scicat(
        cls,
        model: DataFile,
        *,
        source_folder: Union[str, PurePosixPath],
        local_path: Union[str, Path] = None,
    ) -> File:
        return File(
            source_folder=source_folder,
            local_path=local_path,
            model=model,
        )

    @property
    def source_folder(self) -> Optional[str]:
        """Base path on the remote.

        Is ``None`` if the files is not on the remote.
        """
        return self._source_folder

    @source_folder.setter
    def source_folder(self, value: Optional[str]):
        """Set the base path on the remote.

        Must be consistent with the path in the dataset!
        """
        self._source_folder = value

    @property
    def remote_access_path(self) -> Optional[str]:
        """Full path to the file on the remote if it exists."""
        return (
            None
            if self.source_folder is None
            else os.path.join(self.source_folder, self.model.path)
        )

    @property
    def local_path(self) -> Optional[Path]:
        """Path to the local file if it exists."""
        return self._local_path

    @property
    def checksum(self) -> Optional[str]:
        """Checksum of the file."""
        return self._model.chk

    @property
    def size(self) -> int:
        """Size in bytes."""
        return self._model.size

    @property
    def creation_time(self) -> datetime:
        """Data and time when the file was created."""
        return dateutil.parser.parse(self.model.time)

    @property
    def model(self) -> Optional[DataFile]:
        return self._model

    def provide_locally(
        self,
        directory: Union[str, Path],
        *,
        downloader: Downloader,
        checksum_algorithm: Optional[str],
    ) -> Path:
        """Download the file to the local filesystem."""
        directory = Path(directory)
        directory.mkdir(parents=True, exist_ok=True)  # TODO undo if later fails
        local_path = directory / self.model.path
        with downloader.connect_for_download() as con:
            con.download_file(local=local_path, remote=self.remote_access_path)
        self._local_path = local_path
        self._validate_local_file(checksum_algorithm=checksum_algorithm)
        return local_path

    def _validate_local_file(self, *, checksum_algorithm):
        if not self._model.chk:
            get_logger().info(
                "Dataset does not contain a checksum for file '%s'. Skipping check.",
                self._local_path,
            )
            return self._validate_local_file_size()
        if not checksum_algorithm:
            get_logger().warning(
                "File '%s' has a checksum but no algorithm has been defined. "
                "Skipping check. Checksum is %s",
                self._local_path,
                self._model.chk,
            )
            return self._validate_local_file_size()
        actual = checksum_of_file(self._local_path, algorithm=checksum_algorithm)
        if actual != self._model.chk:
            _log_and_raise(
                IntegrityError,
                f"Checksum of file '{self._local_path}' ({actual}) "
                f"does not match checksum stored in dataset "
                f"({self._model.chk}). Using algorithm "
                f"'{checksum_algorithm}'.",
            )

    def _validate_local_file_size(self):
        st = self._local_path.stat()
        actual = st.st_size
        if actual != self._model.size:
            _log_and_raise(
                IntegrityError,
                f"Size of file '{self._local_path}' ({actual}B) does not "
                f"match size stored in dataset ({self._model.size}B)",
            )

    def __repr__(self):
        return (
            f"File(source_folder={self.source_folder}, local_path={self.local_path}, "
            f"model={self.model!r})"
        )


def _log_and_raise(typ, msg):
    get_logger().error(msg)
    raise typ(msg)


def _creation_time_str(st: os.stat_result) -> str:
    """Return the time in UTC when a file was created.

    Uses modification time as SciCat only cares about the latest version of the file
    and not when it was first created on the local system.
    """
    # TODO is this correct on non-linux?
    # TODO is this correct if the file was created in a different timezone (DST)?
    return (
        datetime.fromtimestamp(st.st_mtime)
        .astimezone(timezone.utc)
        .isoformat(timespec="seconds")
    )


def _new_hash(algorithm: str):
    try:
        return hashlib.new(algorithm, usedforsecurity=False)
    except TypeError:
        # Fallback for Python < 3.9
        return hashlib.new(algorithm)


# size based on http://git.savannah.gnu.org/gitweb/?p=coreutils.git;a=blob;f=src/ioblksize.h;h=ed2f4a9c4d77462f357353eb73ee4306c28b37f1;hb=HEAD#l23  # noqa
def checksum_of_file(path: Union[str, Path], *, algorithm: str) -> str:
    chk = _new_hash(algorithm)
    buffer = memoryview(bytearray(128 * 1024))
    with open(path, "rb", buffering=0) as file:
        for n in iter(lambda: file.readinto(buffer), 0):
            chk.update(buffer[:n])
    return chk.hexdigest()


def file_model_from_local_file(
    path: Path,
    *,
    remote_path: PurePosixPath,
    checksum_algorithm: str,
    uid: Optional[str],
    gid: Optional[str],
    perm: Optional[str],
) -> DataFile:
    st = path.stat()
    chk = checksum_of_file(path, algorithm=checksum_algorithm)
    return DataFile(
        path=str(remote_path),
        size=st.st_size,
        time=_creation_time_str(st),
        chk=chk,
        uid=uid,
        gid=gid,
        perm=perm,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen

from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
import re
from string import Template
import sys
from typing import Optional

from pyscicat.model import DerivedDataset, RawDataset

OMIT = (
    "number_of_files",
    "number_of_files_archived",
    "packed_size",
    "pid",
    "size",
    "scientific_metadata",
)


def read_template():
    with open(Path(__file__).parent / "templates/dataset_fields.py.in", "r") as f:
        return Template(f.read())


def camel_case_to_snake_case(s):
    return s[0] + re.sub(r"[A-Z]", lambda m: "_" + m[0].lower(), s[1:])


def normalize_derived_name(name):
    try:
        return {"investigator": "investigator", "type": "dataset_type"}[name]
    except KeyError:
        return camel_case_to_snake_case(name)


def normalize_raw_name(name):
    try:
        return {"principalInvestigator": "investigator", "type": "dataset_type"}[name]
    except KeyError:
        return camel_case_to_snake_case(name)


def join_fields():
    fields = {
        normalize_derived_name(field.name): Field.make(
            field, in_derived=field.name, in_raw=None
        )
        for field in DerivedDataset.__fields__.values()
    }
    for field in RawDataset.__fields__.values():
        name = normalize_raw_name(field.name)
        if name in fields:
            fields[name].name_in_raw = field.name
            # Must be optional if either one is optional
            fields[name].optional |= field.allow_none
        else:
            fields[name] = Field.make(field, in_derived=None, in_raw=field.name)

    # Fields present in only one model must be optional here.
    for field in fields.values():
        if not field.name_in_derived or not field.name_in_raw:
            field.optional = True
    return fields


@dataclass
class Field:
    typ: str
    optional: bool
    name_in_derived: Optional[str]
    name_in_raw: Optional[str]

    @classmethod
    def make(cls, pydantic_field, in_derived, in_raw) -> Field:
        # str of types produces, e.g., <class 'str'> which is not a valid
        # type annotation. So use __name__ in those cases but str(outer_typ)
        # for generic types like typing.List[str].
        outer_typ = pydantic_field.outer_type_
        typ = outer_typ.__name__ if isinstance(outer_typ, type) else str(outer_typ)

        return Field(
            typ=typ,
            optional=pydantic_field.allow_none,
            name_in_derived=in_derived,
            name_in_raw=in_raw,
        )

    @property
    def full_typ(self) -> str:
        return f"Optional[{self.typ}]" if self.optional else self.typ


def generate_omitted_arg_list(fields):
    return (",\n" + " " * 8).join(f"{name}: {fields[name].full_typ}" for name in OMIT)


def generate_omitted_assignment():
    return "\n".join(
        f'        mapped[name_mapping["{name}"]] = {name}' for name in OMIT
    )


def generate_name_in_derived_map():
    indent = " " * 4
    return (
        "{\n"
        + ",\n".join(
            indent + f'"{normalize_derived_name(field.name)}": "{field.name}"'
            for field in DerivedDataset.__fields__.values()
        )
        + ",\n}"
    )


def generate_name_in_raw_map():
    indent = " " * 4
    return (
        "{\n"
        + ",\n".join(
            indent + f'"{normalize_raw_name(field.name)}": "{field.name}"'
            for field in RawDataset.__fields__.values()
        )
        + ",\n}"
    )


def generate_field_spec(name, field):
    default = " = None" if field.optional else ""
    return f"{name}: {field.full_typ}{default}"


def generate_fields_spec(fields):
    return "\n".join(
        f"    {generate_field_spec(name, field)}"
        for name, field in sorted(
            sorted(fields.items(), key=lambda t: t[0]), key=lambda t: t[1].optional
        )
        if name not in OMIT
    )


def generate_dataset():
    template = read_template()
    fields = join_fields()
    src = (
        "# This file was generated by tools/generate_dataset_fields.py\n"
        + template.substitute(
            name_to_derived_map=generate_name_in_derived_map(),
            name_to_raw_map=generate_name_in_raw_map(),
            omitted_fields="(\n" + ",\n".join(f'    "{o}"' for o in OMIT) + "\n)",
            dataclass_fields=generate_fields_spec(fields),
            omitted_arg_list=generate_omitted_arg_list(fields),
            omitted_assignment=generate_omitted_assignment(),
        )
    )
    # TODO conversion to model shows camelCase names in model if validation fails
    return src


def main():
    if len(sys.argv) != 2:
        print("Usage: generate_dataset_fields.py OUTFILE")
        sys.exit(1)
    with open(sys.argv[1], "w") as f:
        f.write(generate_dataset())


if __name__ == "__main__":
    main()



import doctest

import scitacean

import os
import sys

sys.path.insert(0, os.path.abspath("."))

# General information about the project.
project = "scitacean"
copyright = "2022 scitacean contributors"
author = "Scitacean contributors"

html_show_sourcelink = True

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    "nbsphinx",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pyscicat": ("https://scicatproject.github.io/pyscicat/", None),
}

# autodocs includes everything, even irrelevant API internals. autosummary
# looks more suitable in the long run when the API grows.
# For a nice example see how xarray handles its API documentation.
autosummary_generate = True

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = True
napoleon_use_rtype = False
napoleon_preprocess_types = True
napoleon_type_aliases = {}
typehints_defaults = "comma"
typehints_use_rtype = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"
html_sourcelink_suffix = ""  # Avoid .ipynb.txt extensions in sources

# The master toctree document.
master_doc = "index"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = scitacean.__version__
# The full version, including alpha/beta/rc tags.
release = scitacean.__version__

warning_is_error = True

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "left_sidebar_end": ["edit-this-page", "sourcelink"],
    "page_sidebar_items": [],
    "show_nav_level": 1,
    "header_links_before_dropdown": 4,
    "external_links": [
        {"name": "SciCat", "url": "https://scicatproject.github.io/"},
        {
            "name": "SciCat Backend",
            "url": "https://scicatproject.github.io/documentation/",
        },
    ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/SciCatProject/scitacean",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/scitacean/",
            "icon": "fa-brands fa-python",
            "type": "fontawesome",
        },
    ],
}
html_context = {
    "doc_path": "docs",
}
html_sidebars = {
    "**": ["sidebar-nav-bs", "page-toc"],
}

html_title = "Scitacean"
html_logo = "_static/logo.svg"
html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "scitaceandoc"

# -- Options for doctest --------------------------------------------------

# sc.plot returns a Figure object and doctest compares that against the
# output written in the docstring. But we only want to show an image of the
# figure, not its `repr`.
# In addition, there is no need to make plots in doctest as the documentation
# build already tests if those plots can be made.
# So we simply disable plots in doctests.
doctest_global_setup = """
"""

doctest_default_flags = (
    doctest.ELLIPSIS
    | doctest.IGNORE_EXCEPTION_DETAIL
    | doctest.DONT_ACCEPT_TRUE_FOR_1
    | doctest.NORMALIZE_WHITESPACE
)

# -- Options for linkcheck ------------------------------------------------

linkcheck_ignore = [
    # Specific lines in GitHub blobs cannot be found by linkcheck.
    r"https?://github\.com/.*?/blob/[a-f0-9]+/.+?#",
    # Many links for PRs from our release notes. Slow and unlikely to cause issues.
    "https://github.com/SciCatProject/scitacean/pulls/[0-9]+",
]



from dataclasses import dataclass
import pytest
import tempfile
from typing import Dict

from .common import backend


def pytest_addoption(parser):
    parser.addoption(
        "--backend-tests",
        action="store_true",
        default=False,
        help="Select whether to run tests against a real SciCat backend",
    )


@pytest.fixture(scope="module")
def scicat_backend(request, scicat_access):
    """Spin up a SciCat backend and API.

    Does nothing unless the --backend-tests command line option is set.
    """

    if request.config.getoption("--backend-tests"):
        with tempfile.TemporaryDirectory() as temp_dir:
            config_file = backend.configure(temp_dir)
            try:
                backend.start_backend_containers(config_file)
                backend.wait_until_backend_is_live(
                    scicat_access, max_time=20, n_tries=20
                )
                yield True
            finally:
                backend.stop_backend_containers(config_file)
    else:
        yield False


@dataclass
class SciCatAccess:
    url: str
    functional_credentials: Dict[str, str]


@pytest.fixture(scope="module")
def scicat_access():
    return SciCatAccess(
        url="http://localhost/api/v3/",
        functional_credentials={"username": "ingestor", "password": "aman"},
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen

from contextlib import contextmanager
from datetime import datetime, timedelta
from hashlib import md5
from typing import Dict

import dateutil.parser
from pyscicat.model import DataFile
import pytest
from scitacean import File
from scitacean.error import IntegrityError
from scitacean.file import checksum_of_file

from .common.files import make_file


@pytest.fixture
def fake_file(fs):
    return make_file(fs, path="local/dir/events.nxs")


def test_file_from_local(fake_file):
    file = File.from_local(fake_file["path"])

    assert file.source_folder is None
    assert file.remote_access_path is None
    assert file.local_path == fake_file["path"]
    assert file.checksum == fake_file["checksum"]
    assert file.size == fake_file["size"]

    assert file.model.path == str(fake_file["path"])
    assert file.model.size == fake_file["size"]
    assert file.model.chk == fake_file["checksum"]
    assert file.model.uid is None
    assert file.model.gid is None
    assert file.model.perm is None

    assert abs(fake_file["creation_time"] - file.creation_time) < timedelta(seconds=1)
    t = datetime.fromisoformat(file.model.time)
    assert abs(fake_file["creation_time"] - t) < timedelta(seconds=1)


def test_file_from_local_with_base_path(fake_file):
    file = File.from_local(fake_file["path"], base_path="local")

    assert file.source_folder is None
    assert file.remote_access_path is None
    assert file.local_path == fake_file["path"]
    assert file.checksum == fake_file["checksum"]
    assert file.size == fake_file["size"]

    assert file.model.path == "dir/events.nxs"
    assert file.model.size == fake_file["size"]
    assert file.model.chk == fake_file["checksum"]
    assert file.model.uid is None
    assert file.model.gid is None
    assert file.model.perm is None

    assert abs(fake_file["creation_time"] - file.creation_time) < timedelta(seconds=1)
    t = datetime.fromisoformat(file.model.time)
    assert abs(fake_file["creation_time"] - t) < timedelta(seconds=1)


def test_file_from_local_set_remote_path(fake_file):
    file = File.from_local(fake_file["path"], remote_path="remote/location/file.nxs")

    assert file.source_folder is None
    assert file.remote_access_path is None
    assert file.local_path == fake_file["path"]
    assert file.checksum == fake_file["checksum"]
    assert file.size == fake_file["size"]

    assert file.model.path == "remote/location/file.nxs"
    assert file.model.size == fake_file["size"]
    assert file.model.chk == fake_file["checksum"]
    assert file.model.uid is None
    assert file.model.gid is None
    assert file.model.perm is None

    assert abs(fake_file["creation_time"] - file.creation_time) < timedelta(seconds=1)
    t = datetime.fromisoformat(file.model.time)
    assert abs(fake_file["creation_time"] - t) < timedelta(seconds=1)


def test_file_from_local_set_source_folder(fake_file):
    file = File.from_local(fake_file["path"], source_folder="source")

    assert file.source_folder == "source"
    assert file.remote_access_path == "source/local/dir/events.nxs"
    assert file.local_path == fake_file["path"]
    assert file.checksum == fake_file["checksum"]
    assert file.size == fake_file["size"]

    assert file.model.path == "local/dir/events.nxs"
    assert file.model.size == fake_file["size"]
    assert file.model.chk == fake_file["checksum"]
    assert file.model.uid is None
    assert file.model.gid is None
    assert file.model.perm is None

    assert abs(fake_file["creation_time"] - file.creation_time) < timedelta(seconds=1)
    t = datetime.fromisoformat(file.model.time)
    assert abs(fake_file["creation_time"] - t) < timedelta(seconds=1)


def test_file_from_local_set_many_args(fake_file):
    file = File.from_local(
        fake_file["path"],
        base_path="local",
        source_folder="remote",
        remote_uid="user-usy",
        remote_gid="groupy-group",
        remote_perm="wrx",
    )

    assert file.source_folder == "remote"
    assert file.remote_access_path == "remote/dir/events.nxs"
    assert file.local_path == fake_file["path"]
    assert file.checksum == fake_file["checksum"]
    assert file.size == fake_file["size"]

    assert file.model.path == "dir/events.nxs"
    assert file.model.size == fake_file["size"]
    assert file.model.chk == fake_file["checksum"]
    assert file.model.uid == "user-usy"
    assert file.model.gid == "groupy-group"
    assert file.model.perm == "wrx"

    assert abs(fake_file["creation_time"] - file.creation_time) < timedelta(seconds=1)
    t = datetime.fromisoformat(file.model.time)
    assert abs(fake_file["creation_time"] - t) < timedelta(seconds=1)


@pytest.mark.parametrize("alg", ("md5", "sha256", "blake2s"))
def test_file_from_local_select_checksum_algorithm(fake_file, alg):
    file = File.from_local(fake_file["path"], checksum_algorithm=alg)
    expected = checksum_of_file(fake_file["path"], algorithm=alg)
    assert file.checksum == expected
    assert file.model.chk == expected


def test_file_from_scicat():
    model = DataFile(path="dir/image.jpg", size=12345, time="2022-06-22T15:42:53.123Z")
    file = File.from_scicat(model, source_folder="remote/source/folder")

    assert file.source_folder == "remote/source/folder"
    assert file.remote_access_path == "remote/source/folder/dir/image.jpg"
    assert file.local_path is None
    assert file.checksum is None
    assert file.size == 12345
    assert file.creation_time == dateutil.parser.parse("2022-06-22T15:42:53.123Z")

    assert file.model.path == "dir/image.jpg"
    assert file.model.size == 12345
    assert file.model.time == "2022-06-22T15:42:53.123Z"
    assert file.model.chk is None
    assert file.model.uid is None
    assert file.model.gid is None
    assert file.model.perm is None


class FakeDownloader:
    def __init__(self, files: Dict[str, bytes], fs):
        self.files = files
        self.fs = fs

    def download_file(self, *, remote, local):
        self.fs.create_file(local, contents=self.files[remote])

    @contextmanager
    def connect_for_download(self):
        yield self


def test_provide_locally_creates_local_file(fs):
    content = b"random-stuff"
    model = DataFile(
        path="folder/file.txt",
        size=len(content),
        time="2022-06-22T15:42:53+00:00",
        chk=md5(content).hexdigest(),
    )
    file = File.from_scicat(model, source_folder="remote/source/")
    downloader = FakeDownloader(files={"remote/source/folder/file.txt": content}, fs=fs)
    file.provide_locally("./local", downloader=downloader, checksum_algorithm="md5")

    with open("local/folder/file.txt", "rb") as f:
        assert f.read() == content


def test_provide_locally_ignores_checksum_if_alg_is_none(fs):
    content = b"random-stuff"
    # This is not a valid checksum, so it cannot be the checksum of the above.
    bad_checksum = "incorrect-checksum"
    model = DataFile(
        path="folder/file.txt",
        size=len(content),
        time="2022-06-22T15:42:53+00:00",
        chk=bad_checksum,
    )
    file = File.from_scicat(model, source_folder="remote/source/")
    downloader = FakeDownloader(files={"remote/source/folder/file.txt": content}, fs=fs)
    # Does not raise
    file.provide_locally("./local", downloader=downloader, checksum_algorithm=None)


def test_provide_locally_detects_bad_checksum(fs):
    content = b"random-stuff"
    # This is not a valid checksum, so it cannot be the checksum of the above.
    bad_checksum = "incorrect-checksum"
    model = DataFile(
        path="folder/file.txt",
        size=len(content),
        time="2022-06-22T15:42:53+00:00",
        chk=bad_checksum,
    )
    file = File.from_scicat(model, source_folder="remote/source/")
    downloader = FakeDownloader(files={"remote/source/folder/file.txt": content}, fs=fs)
    with pytest.raises(IntegrityError):
        file.provide_locally("./local", downloader=downloader, checksum_algorithm="md5")


def test_provide_locally_detects_bad_size_if_checksum_not_available(fs):
    content = b"random-stuff"
    model = DataFile(
        path="folder/file.txt", size=9999, time="2022-06-22T15:42:53+00:00"
    )
    file = File.from_scicat(model, source_folder="remote/source/")
    downloader = FakeDownloader(files={"remote/source/folder/file.txt": content}, fs=fs)
    with pytest.raises(IntegrityError):
        file.provide_locally("./local", downloader=downloader, checksum_algorithm=None)


def test_provide_locally_detects_bad_size_if_checksum_alg_is_none(fs):
    content = b"random-stuff"
    # This is not a valid checksum, so it cannot be the checksum of the above.
    bad_checksum = "incorrect-checksum"
    model = DataFile(
        path="folder/file.txt",
        size=9999,
        time="2022-06-22T15:42:53+00:00",
        chk=bad_checksum,
    )
    file = File.from_scicat(model, source_folder="remote/source/")
    downloader = FakeDownloader(files={"remote/source/folder/file.txt": content}, fs=fs)
    with pytest.raises(IntegrityError):
        file.provide_locally("./local", downloader=downloader, checksum_algorithm=None)



# This file was generated by tools/generate_dataset_fields.py
import dataclasses
from typing import Any, Dict, Optional, Union
import typing

from pyscicat.model import DatasetType, DerivedDataset, RawDataset

_NAME_IN_DERIVED = {
    "created_by": "createdBy",
    "updated_by": "updatedBy",
    "updated_at": "updatedAt",
    "created_at": "createdAt",
    "owner_group": "ownerGroup",
    "access_groups": "accessGroups",
    "pid": "pid",
    "classification": "classification",
    "contact_email": "contactEmail",
    "creation_time": "creationTime",
    "dataset_name": "datasetName",
    "description": "description",
    "history": "history",
    "instrument_id": "instrumentId",
    "is_published": "isPublished",
    "keywords": "keywords",
    "license": "license",
    "number_of_files": "numberOfFiles",
    "number_of_files_archived": "numberOfFilesArchived",
    "orcid_of_owner": "orcidOfOwner",
    "packed_size": "packedSize",
    "owner": "owner",
    "owner_email": "ownerEmail",
    "shared_with": "sharedWith",
    "size": "size",
    "source_folder": "sourceFolder",
    "source_folder_host": "sourceFolderHost",
    "techniques": "techniques",
    "dataset_type": "type",
    "validation_status": "validationStatus",
    "version": "version",
    "investigator": "investigator",
    "input_datasets": "inputDatasets",
    "used_software": "usedSoftware",
    "job_parameters": "jobParameters",
    "job_log_data": "jobLogData",
    "scientific_metadata": "scientificMetadata",
}
_NAME_IN_RAW = {
    "created_by": "createdBy",
    "updated_by": "updatedBy",
    "updated_at": "updatedAt",
    "created_at": "createdAt",
    "owner_group": "ownerGroup",
    "access_groups": "accessGroups",
    "pid": "pid",
    "classification": "classification",
    "contact_email": "contactEmail",
    "creation_time": "creationTime",
    "dataset_name": "datasetName",
    "description": "description",
    "history": "history",
    "instrument_id": "instrumentId",
    "is_published": "isPublished",
    "keywords": "keywords",
    "license": "license",
    "number_of_files": "numberOfFiles",
    "number_of_files_archived": "numberOfFilesArchived",
    "orcid_of_owner": "orcidOfOwner",
    "packed_size": "packedSize",
    "owner": "owner",
    "owner_email": "ownerEmail",
    "shared_with": "sharedWith",
    "size": "size",
    "source_folder": "sourceFolder",
    "source_folder_host": "sourceFolderHost",
    "techniques": "techniques",
    "dataset_type": "type",
    "validation_status": "validationStatus",
    "version": "version",
    "investigator": "principalInvestigator",
    "creation_location": "creationLocation",
    "data_format": "dataFormat",
    "end_time": "endTime",
    "sample_id": "sampleId",
    "proposal_id": "proposalId",
    "scientific_metadata": "scientificMetadata",
}
_OMITTED_FIELDS = (
    "number_of_files",
    "number_of_files_archived",
    "packed_size",
    "pid",
    "size",
    "scientific_metadata",
)


@dataclasses.dataclass
class DatasetFields:
    access_groups: typing.List[str]
    contact_email: str
    creation_time: str
    dataset_type: DatasetType
    owner: str
    owner_group: str
    source_folder: str
    classification: Optional[str] = None
    created_at: Optional[str] = None
    created_by: Optional[str] = None
    creation_location: Optional[str] = None
    data_format: Optional[str] = None
    dataset_name: Optional[str] = None
    description: Optional[str] = None
    end_time: Optional[str] = None
    history: Optional[typing.List[dict]] = None
    input_datasets: Optional[typing.List[str]] = None
    instrument_id: Optional[str] = None
    investigator: Optional[str] = None
    is_published: Optional[bool] = None
    job_log_data: Optional[str] = None
    job_parameters: Optional[dict] = None
    keywords: Optional[typing.List[str]] = None
    license: Optional[str] = None
    orcid_of_owner: Optional[str] = None
    owner_email: Optional[str] = None
    proposal_id: Optional[str] = None
    sample_id: Optional[str] = None
    shared_with: Optional[typing.List[str]] = None
    source_folder_host: Optional[str] = None
    techniques: Optional[typing.List[dict]] = None
    updated_at: Optional[str] = None
    updated_by: Optional[str] = None
    used_software: Optional[typing.List[str]] = None
    validation_status: Optional[str] = None
    version: Optional[str] = None

    def _map_fields(self, name_mapping, type_name: str) -> Dict[str, Any]:
        mapped = {}
        for name, val in filter(
            lambda t: t[1] is not None, dataclasses.asdict(self).items()
        ):
            try:
                mapped[name_mapping[name]] = val
            except KeyError:
                raise TypeError(
                    f"Field {name} is not allowed" f" in {type_name} datasets"
                ) from None
        return mapped

    def _map_to_model(
        self,
        number_of_files: Optional[int],
        number_of_files_archived: Optional[int],
        packed_size: Optional[int],
        pid: Optional[str],
        size: Optional[int],
        scientific_metadata: Optional[typing.Dict],
    ) -> Union[DerivedDataset, RawDataset]:
        if self.dataset_type == DatasetType.derived:
            name_mapping = _NAME_IN_DERIVED
            type_name = "derived"
            model = DerivedDataset
        else:
            name_mapping = _NAME_IN_RAW
            type_name = "raw"
            model = RawDataset
        mapped = self._map_fields(name_mapping, type_name)
        mapped[name_mapping["number_of_files"]] = number_of_files
        mapped[name_mapping["number_of_files_archived"]] = number_of_files_archived
        mapped[name_mapping["packed_size"]] = packed_size
        mapped[name_mapping["pid"]] = pid
        mapped[name_mapping["size"]] = size
        mapped[name_mapping["scientific_metadata"]] = scientific_metadata
        return model(**mapped)

    @staticmethod
    def _find_field_name_for_model_name(s, mapping):
        for scitacean_name, model_name in mapping.items():
            if model_name == s:
                return scitacean_name
        raise ValueError(f"{s} is not a valid field name")

    @staticmethod
    def _map_model_to_field_dict(model):
        d = {
            DatasetFields._find_field_name_for_model_name(
                name,
                _NAME_IN_DERIVED if model.type == DatasetType.derived else _NAME_IN_RAW,
            ): val
            for name, val in model.dict(exclude_none=True).items()
        }
        for name in _OMITTED_FIELDS:
            d.pop(name, None)
        return d



from __future__ import annotations
from typing import Optional


class PID:
    """Stores the ID of database item.

    The ID is split into a prefix and the main identifier.
    The prefix identifies an instance of SciCat and the main identifier a dataset.

    The two components are merged using a "/", i.e.

    .. code-block:: python

        full_id = PID.prefix + '/' + PID.pid

    Equivalently, ``str`` can be used to construct the full id:

    .. code-block:: python

        full_id = str(PID)
    """

    __slots__ = ("_pid", "_prefix")

    def __init__(self, *, pid: str, prefix: Optional[str] = None):
        """

        Parameters
        ----------
        pid:
            Main part of the ID which uniquely identifies a dataset.
        prefix:
            Identifies the instance of SciCat.
        """
        self._pid = pid
        self._prefix = prefix

    @classmethod
    def parse(cls, x: str) -> PID:
        """Build a PID from a string.

        The string is split at the first "/" to determine
        prefix and main ID.
        This means that it only works if the prefix and main ID do
        not contain any slashes.

        Parameters
        ----------
        x:
            String holding an ID with or without prefix.

        Returns
        -------
        :
            A new PID object constructed from ``x``.
        """
        pieces = x.split("/", 1)
        if len(pieces) == 1:
            return PID(pid=pieces[0], prefix=None)
        return PID(prefix=pieces[0], pid=pieces[1])

    @property
    def pid(self) -> str:
        """Main part of the ID."""
        return self._pid

    @property
    def prefix(self) -> Optional[str]:
        """Prefix part of the ID if there is one."""
        return self._prefix

    @property
    def without_prefix(self) -> PID:
        """Return a new PID with the prefix set to None."""
        return PID(pid=self.pid, prefix=None)

    def __str__(self):
        if self.prefix is not None:
            return self.prefix + "/" + self.pid
        return self.pid

    def __repr__(self):
        return f"PID(prefix={self.prefix}, pid={self.pid})"

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen

from datetime import datetime, timezone
import hashlib
from pathlib import Path
from typing import Any, Dict, Optional, Union


def make_file(
    fs, path=Union[str, Path], contents: Optional[bytes] = None
) -> Dict[str, Any]:
    if contents is None:
        contents = b"a bunch of file contents" * len(str(path))
    path = Path(path)

    checksum = hashlib.new("md5")
    checksum.update(contents)
    checksum = checksum.hexdigest()

    fs.create_file(path, contents=contents)
    # Not exact but should at least be precise to the second
    # and avoids potential difficulties of querying the file system.
    creation_time = datetime.now().astimezone(timezone.utc)

    return dict(
        path=path, creation_time=creation_time, checksum=checksum, size=len(contents)
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen

from typing import Any, Dict, Tuple

from hypothesis import infer, strategies as st
from pyscicat import model as m


def builds_model(
    model, to_infer: Tuple[str, ...], user_kwargs: Dict[str, Any]
) -> st.SearchStrategy:
    """Return a SearchStrategy for a given Pydantic model.

    ``hypothesis.strategies.builds`` cannot be used here because it always sets
    ``Optional`` fields to ``None``,
    see https://github.com/pydantic/pydantic/issues/3126
    As a workaround, ``st.from_type`` does work with ``Optional`` but does
    allow specifying individual fields.
    """
    inferred = {name: infer for name in to_infer}
    customized = {
        name: arg if isinstance(arg, st.SearchStrategy) else st.just(arg)
        for name, arg in user_kwargs.items()
    }
    return st.builds(model, **{**inferred, **customized})


# TODO implement both for model and Dataset class
def datasets(**kwargs) -> st.SearchStrategy[m.Dataset]:
    """Returns a SearchStrategy for a dataset model."""
    # TODO add all other optional fields
    return builds_model(m.Dataset, ("classification", "isPublished"), kwargs)


def derived_datasets(**kwargs) -> st.SearchStrategy[m.DerivedDataset]:
    return builds_model(m.DerivedDataset, ("classification", "isPublished"), kwargs)


def raw_datasets(**kwargs) -> st.SearchStrategy[m.RawDataset]:
    return builds_model(m.RawDataset, ("classification", "isPublished"), kwargs)



"""
Utilities for the documentation.

Should probably not be used externally.
"""

from pyscicat.model import DataFile, OrigDatablock, RawDataset

from ..pid import PID
from .client import FakeClient
from .transfer import FakeFileTransfer


def _create_raw_dataset(client: FakeClient):
    content1 = b"5 4 9 11 15 12 7 6 1"
    content2 = (
        b"INFO Starting measurement\nWARN Detector saturated\nINFO Measurement finished"
    )
    client.file_transfer.files["/hex/ps/thaum/flux.dat"] = content1
    client.file_transfer.files["/hex/ps/thaum/logs/measurement.log"] = content2

    dataset_id = PID(prefix="20.500.12269", pid="72fe3ff6-105b-4c7f-b9d0-073b67c90ec3")
    client.datasets[dataset_id] = RawDataset(
        pid=str(dataset_id),
        datasetName="Thaum flux",
        description="Measured the thaum flux",
        createdBy="Ponder Stibbons",
        updatedBy="anonymous",
        updatedAt="2022-11-01T13:22:08.927Z",
        createdAt="2022-08-17T14:20:23.675Z",
        owner="Ponder Stibbons",
        ownerGroup="uu",
        accessGroups=["faculty"],
        principalInvestigator="Ponder Stibbons",
        contactEmail="p.stibbons@uu.am",
        creationTime="2022-06-29T14:01:05.000Z",
        numberOfFiles=2,
        size=len(content1) + len(content2),
        sourceFolder="/hex/ps/thaum",
        scientificMetadata={
            "data_type": "histogram",
            "temperature": {"value": "123", "unit": "K"},
        },
    )
    client.orig_datablocks[dataset_id] = [
        OrigDatablock(
            id="20.500.12269/02dc390c-811c-4d6a-93bf-9f85a4214ca0",
            datasetId=str(dataset_id),
            size=len(content1) + len(content2),
            ownerGroup="uu",
            accessGroups=["faculty"],
            dataFileList=[
                DataFile(
                    path="flux.dat", size=len(content1), time="2022-08-17T13:54:12Z"
                ),
                DataFile(
                    path="logs/measurement.log",
                    size=len(content2),
                    time="2022-08-17T13:55:21Z",
                ),
            ],
        )
    ]


def setup_fake_client() -> FakeClient:
    client = FakeClient.from_token(  # nosec B106
        url="fake-url.sci/api/v3",
        token="fake-token",
        file_transfer=FakeFileTransfer(fs=None),
    )
    _create_raw_dataset(client)

    return client



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen
"""Fake client for testing."""

from __future__ import annotations

import functools
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional, Union
import uuid

from pyscicat import model
import pyscicat.client

from ..client import Client, ScicatClient
from ..dataset import Dataset
from ..pid import PID
from ..typing import FileTransfer


def _conditionally_disabled(func):
    @functools.wraps(func)
    def impl(self, *args, **kwargs):
        if (exc := self.main.disabled.get(func.__name__)) is not None:
            raise exc
        return func(self, *args, **kwargs)

    return impl


# Inherits from Client to satisfy type hints.
class FakeClient(Client):
    """Mimics a client without accessing the internet.

    This class is a stand in for :class:`scitacean.Client` to emulate downloading and
    uploading of datasets without actually accessing a SciCat server.

    Since this client fakes communication with SciCat, it stores raw
    `pydantic <https://pydantic-docs.helpmanual.io/>`_ models, namely

    - ``FakeClient.datasets``:
            :class:`dict` of :class:`pydantic.model.DerivedDataset` and
            :class:`pydantic.model.RawDataset`,
            indexed by dataset ID.
    - ``FakeClient.orig_datablocks``:
            :class:`dict` of lists of :class:`pydantic.model.OrigDatablock`,
            indexed by the *dataset* ID.

    Individual functions can be disabled (that is made to raise an exception)
    using the ``disabled`` argument of the initializer.

    Important
    ---------
    ``FakeClient`` does not behave exactly like a ``Client`` connected to a real
    server as that would require reimplementing a large part of the SciCat backend.
    You should thus always test your code against a (potentially locally deployed)
    real server.

    In particular, do not rely on specific error messages or the detailed settings
    in the datasets returned by ``FakeClient.upload_new_dataset_now``!

    Examples
    --------
    Set up a fake client for download:

    .. code-block:: python

        client = FakeClient()
        pid = PID(prefix="sample.prefix", pid="1234-5678-abcd")
        client.datasets[pid] = pyscicat.model.DerivedDataset(...)
        client.orig_datablocks[pid] = [pyscicat.model.OrigDatablock(
            datasetID=str(pid),
            ...
        )]
        client.get_dataset(pid)

    Use to upload a dataset:

    .. code-block:: python

        client = FakeClient(file_transfer=FakeFileTransfer())
        finalized = client.upload_new_dataset_now(dset)

        # contains new dataset and datablock:
        client.datasets[finalized.pid]
        client.orig_datablocks[finalized.pid]

    Disable a method:

    .. code-block:: python

        client = FakeClient(
            disable={"create_dataset_model": RuntimeError("Upload failed")})
        # raises RuntimeError("Upload failed"):
        client.upload_new_dataset_now(dset)

    See Also
    --------
    scitacean.testing.transfer.FakeFileTransfer: Fake file up-/downloads
        without a real server.
    """

    def __init__(
        self,
        *,
        file_transfer: Optional[FileTransfer] = None,
        disable: Optional[Dict[str, Exception]] = None,
    ):
        """Initialize a fake client with empty dataset storage.

        Parameters
        ----------
        file_transfer:
            Typically :class:`scitacean.testing.transfer.FakeFileTransfer`
            but may be a real file transfer.
        disable:
            dict function names to exceptions. Functions listed here raise
            the given exception when called and do nothing else.
        """
        # Normally, client must not be None, but the fake must never
        # call it, so setting it to None serves as an extra safeguard.
        super().__init__(client=None, file_transfer=file_transfer)  # noqa

        self._scicat_client = FakeScicatClient(self)
        self.disabled = {} if disable is None else dict(disable)
        self.datasets: Dict[PID, Union[model.DerivedDataset, model.RawDataset]] = {}
        self.orig_datablocks: Dict[PID, List[model.OrigDatablock]] = {}

    @classmethod
    def from_token(
        cls, *, url: str, token: str, file_transfer: Optional[FileTransfer] = None
    ) -> FakeClient:
        """Create a new fake client.

        All arguments except ``file_Transfer`` are ignored.
        """
        return FakeClient(file_transfer=file_transfer)

    @classmethod
    def from_credentials(
        cls,
        *,
        url: str,
        username: str,
        password: str,
        file_transfer: Optional[FileTransfer] = None,
    ) -> FakeClient:
        """Create a new fake client.

        All arguments except ``file_Transfer`` are ignored.
        """
        return FakeClient(file_transfer=file_transfer)

    def get_dataset(self, pid: Union[PID, str]) -> Dataset:
        """Return a dataset from the client's internal storage."""
        return Dataset.from_models(
            dataset_model=self.scicat.get_dataset_model(str(pid)),
            orig_datablock_models=self.scicat.get_orig_datablocks(str(pid)),
        )

    @property
    def scicat(self) -> FakeScicatClient:
        """Client for lower level SciCat communication."""
        return self._scicat_client

    def download_file(self, *, remote: Union[str, Path], local: Union[str, Path]):
        """Download a file using the client's file_transfer."""
        if self.file_transfer is None:
            raise RuntimeError(
                f"No file transfer handler specified, cannot download file {remote}"
            )
        with self.file_transfer.connect_for_download() as con:
            con.download_file(remote=remote, local=local)

    def upload_file(
        self, *, dataset_id: PID, remote: Union[str, Path], local: Union[str, Path]
    ) -> str:
        """Upload a file using the client's file_transfer."""
        if self.file_transfer is None:
            raise RuntimeError(
                f"No file transfer handler specified, cannot upload file {local}"
            )
        with self.file_transfer.connect_for_upload(dataset_id) as con:
            return con.upload_file(remote=remote, local=local)


class FakeScicatClient(ScicatClient):
    """Mimics a ScicatClient, to be used by FakeClient."""

    def __init__(self, main_client):
        super().__init__(None)  # noqa
        self.main = main_client

    @_conditionally_disabled
    def get_dataset_model(
        self, pid: str
    ) -> Union[model.DerivedDataset, model.RawDataset]:
        try:
            return self.main.datasets[pid]
        except KeyError:
            raise pyscicat.client.ScicatCommError(
                f"Unable to retrieve dataset {pid}"
            ) from None

    @_conditionally_disabled
    def get_orig_datablocks(self, pid: str) -> List[model.OrigDatablock]:
        try:
            return self.main.orig_datablocks[pid]
        except KeyError:
            raise pyscicat.client.ScicatCommError(
                f"Unable to retrieve orig datablock for dataset {pid}"
            ) from None

    @_conditionally_disabled
    def create_dataset_model(self, dset: model.Dataset) -> str:
        pid = PID(
            pid=dset.pid if dset.pid is not None else str(uuid.uuid4()),
            prefix="PID.SAMPLE.PREFIX",
        )
        if pid in self.main.datasets:
            raise pyscicat.client.ScicatCommError(f"Dataset id already exists: {pid}")
        self.main.datasets[pid] = deepcopy(dset)
        self.main.datasets[pid].pid = str(pid)
        return str(pid)

    @_conditionally_disabled
    def create_orig_datablock(self, dblock: model.OrigDatablock):
        dataset_id = PID.parse(dblock.datasetId)
        if dataset_id not in self.main.datasets:
            raise pyscicat.client.ScicatCommError(f"No dataset with id {dataset_id}")
        self.main.orig_datablocks.setdefault(dataset_id, []).append(dblock)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen

from contextlib import contextmanager
from getpass import getpass
import os
from pathlib import Path
from typing import Union, Tuple, Optional

# These are quite heavy dependencies.
# It would be great if we could do without them in the long run.
# Note that invoke and paramiko are dependencies of fabric.
from fabric import Connection
from invoke.exceptions import UnexpectedExit
from paramiko.ssh_exception import AuthenticationException, PasswordRequiredException

from ..logging import get_logger
from ..pid import PID


# TODO process multiple files together
# TODO pass pid in put/revert?
#      downloading does not need a pid, so it should not be required in the constructor/
# TODO cache download (maybe using pooch)
class ESSTestFileTransfer:
    """Transfer files to/from test directory at ESS using SSH."""

    def __init__(
        self,
        *,
        host: str = "login.esss.dk",
        port: Optional[int] = None,
        remote_base_path: str = "/mnt/groupdata/scicat/upload",
    ):
        self._host = host
        self._port = port
        self._remote_base_path = remote_base_path

    @contextmanager
    def connect_for_download(self):
        with _connect(self._host, self._port) as con:
            yield _ESSDownloadConnection(connection=con)

    @contextmanager
    def connect_for_upload(self, dataset_id: PID):
        with _connect(self._host, self._port) as con:
            yield _ESSUploadConnection(
                connection=con,
                dataset_id=dataset_id,
                remote_base_path=self._remote_base_path,
            )


class _ESSDownloadConnection:
    def __init__(self, *, connection: Connection):
        self._connection = connection

    def download_file(self, *, remote: Union[str, Path], local: Union[str, Path]):
        """Download a file from the given remote path."""
        get_logger().info(
            "Downloading file %s from host %s to %s",
            remote,
            self._connection.host,
            local,
        )
        self._connection.get(remote=str(remote), local=str(local))


class _ESSUploadConnection:
    def __init__(
        self, *, connection: Connection, dataset_id: PID, remote_base_path: str
    ):
        self._connection = connection
        self._dataset_id = dataset_id
        self._remote_base_path = remote_base_path

    @property
    def source_dir(self) -> str:
        return os.path.join(self._remote_base_path, self._dataset_id.pid)

    def remote_path(self, filename) -> str:
        return os.path.join(self.source_dir, filename)

    def upload_file(self, *, remote: Union[str, Path], local: Union[str, Path]) -> str:
        """Upload a file to the remote staging folder."""
        remote_path = self.remote_path(remote)
        get_logger().info(
            "Uploading file %s to %s on host %s", local, remote, self._connection.host
        )

        self._connection.run(f"mkdir -p {self.source_dir}", hide=True)
        try:
            self._connection.put(
                remote=str(remote_path),
                local=str(local),
            )
        except Exception as exc:
            if _folder_is_empty(self._connection, self.source_dir):
                self._connection.run(f"rm -r {self.source_dir}", hide=True)
            raise exc
        return remote_path

    def revert_upload(self, *, remote: Union[str, Path], local: Union[str, Path] = ""):
        """Remove an uploaded file from the remote staging folder."""
        remote_path = self.remote_path(remote)
        get_logger().info(
            "Reverting upload of file %s to %s on host %s",
            local,
            remote_path,
            self._connection.host,
        )

        try:
            self._connection.run(f"rm {remote_path}", hide=True)
        except UnexpectedExit as exc:
            get_logger().warning(
                "Error reverting file %s:\n%s", remote_path, exc.result
            )
            return

        if _folder_is_empty(self._connection, self.source_dir):
            try:
                get_logger().info(
                    "Removing empty remote directory %s on host %s",
                    self.source_dir,
                    self._connection.host,
                )
                self._connection.run(f"rm -r {self.source_dir}", hide=True)
            except UnexpectedExit as exc:
                get_logger().warning(
                    "Failed to remove empty remote directory %s on host:\n%s",
                    self.source_dir,
                    self._connection.host,
                    exc.result,
                )


def _ask_for_key_passphrase() -> str:
    return getpass("The private key is encrypted, enter passphrase: ")


def _ask_for_credentials(host) -> Tuple[str, str]:
    print(f"You need to authenticate to access {host}")
    username = input("Username: ")
    password = getpass("Password: ")
    return username, password


@contextmanager
def _generic_connect(host, port, **kwargs):
    with Connection(host=host, port=port, **kwargs) as con:
        con.open()
        yield con


@contextmanager
def _unauthenticated_connect(host, port):
    with _generic_connect(host, port) as con:
        yield con


@contextmanager
def _authenticated_connect(host, port, exc: AuthenticationException):
    # TODO fail fast if output going to file
    if isinstance(exc, PasswordRequiredException) and "encrypted" in exc.args[0]:
        # TODO does not work anymore, exception is always AuthenticationException
        with _generic_connect(
            host=host,
            port=port,
            connect_kwargs={"passphrase": _ask_for_key_passphrase()},
        ) as con:
            yield con
    else:
        username, password = _ask_for_credentials(host)
        with _generic_connect(
            host=host, port=port, user=username, connect_kwargs={"password": password}
        ) as con:
            yield con


@contextmanager
def _connect(host, port):
    # Catch all exceptions and remove traceback up to this point.
    # We pass secrets as arguments to functions called in this block and those
    # can be leaked through exception handlers.
    try:
        try:
            with _unauthenticated_connect(host, port) as con:
                yield con
        except AuthenticationException as exc:
            with _authenticated_connect(host, port, exc) as con:
                yield con
    except Exception as exc:
        # TODO dedicated exception type?
        raise type(exc)(exc.args) from None
    except BaseException as exc:
        raise type(exc)(exc.args) from None


def _remote_path_exists(con: Connection, path: str) -> bool:
    return con.run(f"stat {path}", hide=True, warn=True).exited == 0


def _folder_is_empty(con, path) -> bool:
    try:
        return con.run(f"ls {path}", hide=True).stdout == ""
    except UnexpectedExit:
        return False  # no further processing is needed in this case



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen

from datetime import datetime, timedelta
from pathlib import Path

from hypothesis import given, settings
import pytest
from pyscicat import model

from scitacean import Dataset
from scitacean.testing import strategies as sst

from .common.files import make_file


@pytest.fixture
def ownable():
    return model.Ownable(ownerGroup="uu", accessGroups=["group1", "2nd_group"])


@pytest.fixture
def derived_dataset_model(ownable):
    return model.DerivedDataset(
        pid="01.432.56789/12345678-abcd-0987-0123456789ab",
        owner="PonderStibbons",
        investigator="Ridcully",
        contactEmail="p.stibbons@uu.am",
        sourceFolder="/hex/source123",
        size=168456,
        numberOfFiles=2,
        creationTime="2011-08-24T12:34:56Z",
        datasetName="Data A38",
        inputDatasets=[],
        usedSoftware=["EasyScience"],
        scientificMetadata={
            "data_type": "event data",
            "temperature": {"value": "123", "unit": "K"},
            "weight": {"value": "42", "unit": "mg"},
        },
        **ownable.dict(),
    )


@pytest.fixture
def orig_datablock_model(derived_dataset_model, ownable):
    return model.OrigDatablock(
        id="fedcba98-5647-a3b2-a0b1c2d3e4f567",
        size=168456,
        datasetId=derived_dataset_model.pid,
        dataFileList=[
            model.DataFile(path="file1.nxs", size=123456, time="2022-02-02T12:34:56Z"),
            model.DataFile(
                path="sub/file2.nxs",
                size=45000,
                time="2022-02-02T12:54:32Z",
            ),
        ],
        **ownable.dict(),
    )


@settings(max_examples=10)
@given(
    sst.derived_datasets(owner="Nanny Ogg", usedSoftware=["Magick"], isPublished=True)
)
def test_can_get_dataset_properties_derived(derived_dataset):
    dset = Dataset.new(model=derived_dataset)
    assert dset.owner == "Nanny Ogg"
    assert dset.used_software == ["Magick"]
    assert dset.data_format is None  # only used for raw datasets
    assert dset.is_published
    assert dset.license == derived_dataset.license


@settings(max_examples=10)
@given(sst.raw_datasets(owner="Nanny Ogg", dataFormat="SpellBook", isPublished=True))
def test_can_get_dataset_properties_raw(raw_dataset):
    dset = Dataset.new(model=raw_dataset)
    assert dset.owner == "Nanny Ogg"
    assert dset.used_software is None  # only used for derived datasets
    assert dset.data_format == "SpellBook"
    assert dset.is_published
    assert dset.license == raw_dataset.license


@settings(max_examples=10)
@given(sst.derived_datasets())
def test_can_set_dataset_properties(derived_dataset):
    dset = Dataset.new(model=derived_dataset)
    dset.owner = "Esmeralda Weatherwax"
    dset.source_folder = "/lancre/coven"
    assert dset.owner == "Esmeralda Weatherwax"
    assert dset.source_folder == "/lancre/coven"


@settings(max_examples=10)
@given(sst.derived_datasets())
def test_cannot_access_some_dataset_properties(derived_dataset):
    # Some of DataFile's properties are exposed in a read only way.
    dset = Dataset.new(model=derived_dataset)
    with pytest.raises(AttributeError):
        dset.size = 1  # noqa
    with pytest.raises(AttributeError):
        dset.files = [model.DataFile(path="path", size=4)]  # noqa


@settings(max_examples=10)
@given(sst.derived_datasets(datasetName="Data A38", owner="Ponder Stibbons"))
def test_new_kwargs_and_model(derived_dataset):
    dset = Dataset.new(
        model=derived_dataset, owner="The Librarian", license="NO-TOUCHY"
    )
    assert dset.dataset_name == "Data A38"
    assert dset.owner == "The Librarian"
    assert dset.license == "NO-TOUCHY"


def test_new_only_kwargs():
    dset = Dataset.new(
        dataset_type="derived",
        owner="The Librarian",
        contact_email="p.stibbons@uu.am",
        source_folder="/hex/source123",
        creation_time="2011-08-24T12:34:56Z",
        dataset_name="Data A38",
        owner_group="uu",
        access_groups=["group1", "2nd_group"],
    )
    assert dset.dataset_name == "Data A38"
    assert dset.owner == "The Librarian"


def test_can_access_scientific_metadata(derived_dataset_model):
    dset = Dataset.new(model=derived_dataset_model)
    assert dset.meta["temperature"] == {"value": "123", "unit": "K"}
    assert dset.meta["data_type"] == "event data"


def test_can_add_more_metadata_in_new(derived_dataset_model):
    dset = Dataset.new(model=derived_dataset_model, meta={"mood": "grumpy"})
    assert dset.meta["temperature"] == {"value": "123", "unit": "K"}
    assert dset.meta["data_type"] == "event data"
    assert dset.meta["mood"] == "grumpy"


def test_only_explicit_metadata_in_new(derived_dataset_model):
    derived_dataset_model.scientificMetadata = None
    dset = Dataset.new(model=derived_dataset_model, meta={"mood": "grumpy"})
    assert "temperature" not in dset.meta
    assert "data_type" not in dset.meta
    assert dset.meta["mood"] == "grumpy"


@settings(max_examples=10)
@given(sst.derived_datasets())
def test_new_dataset_has_no_files(derived_dataset):
    dset = Dataset.new(model=derived_dataset)
    assert len(dset.files) == 0
    assert dset.size == 0


def test_add_local_file_to_new_dataset(derived_dataset_model, fs):
    f = make_file(fs, "local/folder/data.dat")

    dset = Dataset.new(model=derived_dataset_model)
    dset.add_local_files("local/folder/data.dat")

    assert len(dset.files) == 1
    assert dset.size == f["size"]

    assert dset.files[0].source_folder is None
    assert dset.files[0].remote_access_path is None
    assert dset.files[0].local_path == Path("local/folder/data.dat")
    assert dset.files[0].size == f["size"]

    assert abs(f["creation_time"] - dset.files[0].creation_time) < timedelta(seconds=1)
    t = datetime.fromisoformat(dset.files[0].model.time)
    assert abs(f["creation_time"] - t) < timedelta(seconds=1)


def test_add_multiple_local_files_to_new_dataset(derived_dataset_model, fs):
    f0 = make_file(fs, "common/location1/data.dat")
    f1 = make_file(fs, "common/song.mp3")

    dset = Dataset.new(model=derived_dataset_model)
    dset.add_local_files("common/location1/data.dat", "common/song.mp3")

    assert len(dset.files) == 2
    assert dset.size == f0["size"] + f1["size"]

    assert dset.files[0].source_folder is None
    assert dset.files[0].remote_access_path is None
    assert dset.files[0].local_path == Path("common/location1/data.dat")
    assert dset.files[0].size == f0["size"]
    assert dset.files[0].model.path == "common/location1/data.dat"

    assert dset.files[1].source_folder is None
    assert dset.files[1].remote_access_path is None
    assert dset.files[1].local_path == Path("common/song.mp3")
    assert dset.files[1].size == f1["size"]
    assert dset.files[1].model.path == "common/song.mp3"


def test_add_multiple_local_files_to_new_dataset_with_base_path(
    derived_dataset_model, fs
):
    f0 = make_file(fs, "common/location1/data.dat")
    f1 = make_file(fs, "common/song.mp3")

    dset = Dataset.new(model=derived_dataset_model)
    dset.add_local_files(
        "common/location1/data.dat", "common/song.mp3", base_path="common"
    )

    assert len(dset.files) == 2
    assert dset.size == f0["size"] + f1["size"]

    assert dset.files[0].source_folder is None
    assert dset.files[0].remote_access_path is None
    assert dset.files[0].local_path == Path("common/location1/data.dat")
    assert dset.files[0].size == f0["size"]
    assert dset.files[0].model.path == "location1/data.dat"

    assert dset.files[1].source_folder is None
    assert dset.files[1].remote_access_path is None
    assert dset.files[1].local_path == Path("common/song.mp3")
    assert dset.files[1].size == f1["size"]
    assert dset.files[1].model.path == "song.mp3"


def test_dataset_from_models(derived_dataset_model, orig_datablock_model):
    dset = Dataset.from_models(
        dataset_model=derived_dataset_model,
        orig_datablock_models=[orig_datablock_model],
    )

    assert dset.source_folder == "/hex/source123"
    assert dset.creation_time == "2011-08-24T12:34:56Z"
    assert dset.access_groups == ["group1", "2nd_group"]
    assert dset.meta["temperature"] == {"value": "123", "unit": "K"}
    assert dset.meta["data_type"] == "event data"

    assert dset.files[0].source_folder == "/hex/source123"
    assert dset.files[0].remote_access_path == "/hex/source123/file1.nxs"
    assert dset.files[0].local_path is None
    assert dset.files[1].source_folder == "/hex/source123"
    assert dset.files[1].remote_access_path == "/hex/source123/sub/file2.nxs"
    assert dset.files[1].local_path is None


def test_dataset_from_models_multi_block_not_supported(
    derived_dataset_model, orig_datablock_model
):
    with pytest.raises(NotImplementedError):
        Dataset.from_models(
            dataset_model=derived_dataset_model,
            orig_datablock_models=[orig_datablock_model, orig_datablock_model],
        )


@settings(max_examples=10)
@given(sst.derived_datasets())
def test_make_scicat_models_creates_correct_dataset(derived_dataset):
    derived_dataset.pid = "some-pid"
    # The default arg of isPublished messes with model creation in Dataset
    derived_dataset.isPublished = False
    dset = Dataset.new(model=derived_dataset)
    mod = dset.make_scicat_models().dataset
    assert mod == derived_dataset


@settings(max_examples=10)
@given(sst.derived_datasets())
def test_make_scicat_models_datablock_without_files(derived_dataset):
    derived_dataset.pid = "some-pid"
    dset = Dataset.new(model=derived_dataset)
    mod = dset.make_scicat_models().datablock
    assert mod.id is None
    assert mod.size == 0
    assert len(mod.dataFileList) == 0
    assert mod.datasetId == "some-pid"


def test_make_scicat_models_datablock_with_files(derived_dataset_model, fs):
    f0 = make_file(fs, "base/folder/events.nxs")
    f1 = make_file(fs, "base/folder/log/wtf.log")

    dset = Dataset.new(model=derived_dataset_model)
    dset.add_local_files(
        "base/folder/events.nxs", "base/folder/log/wtf.log", base_path="base/folder"
    )

    mod = dset.make_scicat_models().datablock

    assert mod.id is None
    assert mod.size == f0["size"] + f1["size"]
    assert len(mod.dataFileList) == 2
    assert mod.datasetId == str(dset.pid)

    assert mod.dataFileList[0].path == "events.nxs"
    assert mod.dataFileList[0].size == f0["size"]
    assert mod.dataFileList[1].path == "log/wtf.log"
    assert mod.dataFileList[1].size == f1["size"]


@settings(max_examples=10)
@given(sst.derived_datasets(pid=None))
def test_make_scicat_models_requires_pid(derived_dataset):
    dset = Dataset.new(model=derived_dataset)
    with pytest.raises(ValueError):
        _ = dset.make_scicat_models()


# inputDatasets is only allowed in derived datasets
@settings(max_examples=10)
@given(sst.derived_datasets(pid="0987-poi", inputDatasets=["abcd-1234"]))
def test_make_derived_scicat_models_fails_with_raw_dataset_type(derived_dataset):
    dset = Dataset.new(model=derived_dataset)
    dset.dataset_type = model.DatasetType.raw
    with pytest.raises(TypeError):
        _ = dset.make_scicat_models()


# sampleId is only allowed in raw datasets
@settings(max_examples=10)
@given(sst.raw_datasets(pid="abcd-1234", sampleId="sample-xyz"))
def test_make_raw_scicat_models_fails_with_derived_dataset_type(raw_dataset):
    dset = Dataset.new(model=raw_dataset)
    dset.dataset_type = model.DatasetType.derived
    with pytest.raises(TypeError):
        _ = dset.make_scicat_models()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen

from contextlib import contextmanager
from copy import deepcopy

from pyscicat.client import ScicatCommError
from pyscicat import model
import pytest
from scitacean import Dataset
from scitacean.testing.client import FakeClient
from scitacean.testing.transfer import FakeFileTransfer

from .common.files import make_file

# TODO source_folder changed / populated in file after upload


@pytest.fixture
def ownable():
    return model.Ownable(ownerGroup="uu", accessGroups=["group1", "2nd_group"])


@pytest.fixture
def derived_dataset_model(ownable):
    return model.DerivedDataset(
        pid="01.432.56789/12345678-abcd-0987-0123456789ab",
        owner="PonderStibbons",
        investigator="Ridcully",
        contactEmail="p.stibbons@uu.am",
        sourceFolder="/hex/source123",
        size=0,
        numberOfFiles=0,
        creationTime="2011-08-24T12:34:56Z",
        datasetName="Data A38",
        inputDatasets=[],
        usedSoftware=["EasyScience"],
        scientificMetadata={
            "data_type": "event data",
            "temperature": {"value": "123", "unit": "K"},
            "weight": {"value": "42", "unit": "mg"},
        },
        **ownable.dict(),
    )


@pytest.fixture
def client(fs):
    return FakeClient(file_transfer=FakeFileTransfer(fs=fs, files={}, reverted={}))


@pytest.fixture
def dataset(derived_dataset_model, fs):
    make_file(fs, path="file.nxs", contents=b"contents of file.nxs")
    make_file(fs, path="the_log_file.log", contents=b"this is a log file")
    dset = Dataset.new(model=derived_dataset_model)
    dset.add_local_files("file.nxs", "the_log_file.log")
    return dset


def test_upload_assigns_fixed_fields(client, dataset):
    expected = deepcopy(dataset)
    finalized = client.upload_new_dataset_now(dataset)
    expected.pid = finalized.pid

    with client.file_transfer.connect_for_upload(finalized.pid) as con:
        source_dir = con.source_dir
    expected.source_folder = source_dir
    assert finalized == expected


def test_upload_creates_dataset_and_datablock(client, dataset):
    finalized = client.upload_new_dataset_now(dataset)
    assert client.datasets[finalized.pid] == finalized.make_scicat_models().dataset
    assert client.orig_datablocks[finalized.pid] == [
        finalized.make_scicat_models().datablock
    ]


def test_upload_uploads_files_to_source_folder(client, dataset):
    finalized = client.upload_new_dataset_now(dataset)

    with client.file_transfer.connect_for_upload(finalized.pid) as con:
        source_dir = con.source_dir
    assert (
        client.file_transfer.files[source_dir + "file.nxs"] == b"contents of file.nxs"
    )
    assert (
        client.file_transfer.files[source_dir + "the_log_file.log"]
        == b"this is a log file"
    )


def test_upload_does_not_create_dataset_if_file_upload_fails(dataset, fs):
    class RaisingUpload(FakeFileTransfer):
        source_dir = "/"

        def upload_file(self, *, local, remote):
            raise RuntimeError("Fake upload failure")

        @contextmanager
        def connect_for_upload(self, pid):
            yield self

    client = FakeClient(file_transfer=RaisingUpload(fs=fs))

    with pytest.raises(RuntimeError, match="Fake upload failure"):
        client.upload_new_dataset_now(dataset)

    assert dataset.pid not in client.datasets
    assert dataset.pid not in client.orig_datablocks


def test_upload_cleans_up_files_if_dataset_ingestion_fails(dataset, fs):
    client = FakeClient(
        disable={"create_dataset_model": ScicatCommError("Ingestion failed")},
        file_transfer=FakeFileTransfer(fs=fs),
    )
    with pytest.raises(ScicatCommError):
        client.upload_new_dataset_now(dataset)

    assert not client.file_transfer.files


def test_failed_datablock_upload_does_not_revert(dataset, fs):
    client = FakeClient(
        disable={"create_orig_datablock": ScicatCommError("Ingestion failed")},
        file_transfer=FakeFileTransfer(fs=fs),
    )
    with pytest.raises(RuntimeError):
        client.upload_new_dataset_now(dataset)

    uploaded_dset = next(iter(client.datasets.values()))
    dataset.pid = uploaded_dset.pid
    dataset.source_folder = uploaded_dset.sourceFolder
    assert uploaded_dset == dataset.make_scicat_models().dataset
    assert client.file_transfer.files
    assert not client.file_transfer.reverted



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen
from copy import deepcopy

import dateutil.parser
import pyscicat.client
from pyscicat.model import DataFile, DatasetType, DerivedDataset, OrigDatablock
import pytest

from scitacean.testing.client import FakeClient
from scitacean import Client

from . import data


@pytest.fixture(scope="module")
def derived_dataset():
    return data.as_dataset_model(data.load_datasets()[0])


@pytest.fixture
def orig_datablock(derived_dataset):
    return data.as_orig_datablock_model(data.load_orig_datablocks()[0])


def make_fake_client(dataset, datablock):
    client = FakeClient(file_transfer=None)
    client.datasets[dataset.pid] = dataset
    client.orig_datablocks[dataset.pid] = [datablock]
    return client


@pytest.fixture(params=["real", "fake"])
def client(
    request,
    derived_dataset,
    orig_datablock,
    scicat_access,
    scicat_backend,
):
    if request.param == "fake":
        return make_fake_client(derived_dataset, orig_datablock)
    if request.param == "real":
        if not request.config.getoption("--backend-tests"):
            # The backend only exists if this option is set.
            pytest.skip(
                "Tests against a real backend are disabled, "
                "use --backend-tests to enable them"
            )
        return Client.from_credentials(
            url=scicat_access.url, **scicat_access.functional_credentials
        )


@pytest.fixture
def scicat_client(client):
    return client.scicat


def test_from_token_fake():
    # This should not call the API
    assert isinstance(
        FakeClient.from_token(url="some.url/api/v3", token="a-token"), FakeClient
    )


def test_from_credentials_fake():
    # This should not call the API
    assert isinstance(
        FakeClient.from_credentials(
            url="some.url/api/v3", username="someone", password="the-fake-does-not-care"
        ),
        FakeClient,
    )


def test_from_credentials_real(scicat_access, scicat_backend):
    if not scicat_backend:
        pytest.skip("No backend")
    Client.from_credentials(
        url=scicat_access.url, **scicat_access.functional_credentials
    )


def test_get_dataset_model(scicat_client, derived_dataset):
    dset = scicat_client.get_dataset_model(derived_dataset.pid)
    assert dset == derived_dataset


def test_get_dataset_model_bad_id(scicat_client):
    with pytest.raises(pyscicat.client.ScicatCommError):
        scicat_client.get_dataset_model("bad-pid")


def test_get_orig_datablock(scicat_client, orig_datablock):
    dblock = scicat_client.get_orig_datablocks(orig_datablock.datasetId)
    assert [orig_datablock] == dblock


def test_get_orig_datablock_bad_id(scicat_client):
    with pytest.raises(pyscicat.client.ScicatCommError):
        scicat_client.get_orig_datablocks("bollocks")


@pytest.mark.parametrize("pid", ("PID.SAMPLE.PREFIX/jeck", "kamelle", None))
def test_create_dataset_model(pid, scicat_client):
    dset = DerivedDataset(
        pid=pid,
        contactEmail="black.foess@dom.k",
        creationTime="1995-11-11T11:11:11.000Z",
        owner="bfoess",
        investigator="bfoess",
        sourceFolder="/dom/platt",
        type=DatasetType.derived,
        inputDatasets=[],
        usedSoftware=[],
        ownerGroup="bfoess",
        accessGroups=["koelle"],
    )
    full_pid = scicat_client.create_dataset_model(dset)
    dset.pid = full_pid
    downloaded = scicat_client.get_dataset_model(full_pid)
    for key, expected in dset.dict(exclude_none=True).items():
        # The database populates a number of fields that are None in dset.
        # But we don't want to test those here as we don't want to test the database.
        assert expected == downloaded.dict()[key], f"key = {key}"


def test_create_dataset_model_id_clash(scicat_client, derived_dataset):
    dset = deepcopy(derived_dataset)
    dset.pid = dset.pid.split("/")[1]
    dset.owner = "a new owner to trigger a failure"
    with pytest.raises(pyscicat.client.ScicatCommError):
        scicat_client.create_dataset_model(dset)


def test_create_first_orig_datablock(scicat_client, derived_dataset):
    dset = deepcopy(derived_dataset)
    dset.pid = "new-dataset-id"
    dset.pid = scicat_client.create_dataset_model(dset)

    dblock = OrigDatablock(
        id="PID.SAMPLE.PREFIX/new-datablock-id",
        size=9235,
        dataFileList=[
            DataFile(path="data.nxs", size=9235, time="2023-08-18T13:52:33.000Z")
        ],
        datasetId=dset.pid,
        ownerGroup="uu",
        accessGroups=["group1", "2nd_group"],
    )
    scicat_client.create_orig_datablock(dblock)
    downloaded = scicat_client.get_orig_datablocks(dset.pid)
    assert len(downloaded) == 1
    downloaded = downloaded[0]
    for key, expected in dblock.dict(exclude_none=True).items():
        # The database populates a number of fields that are None in dset.
        # But we don't want to test those here as we don't want to test the database.
        if key != "dataFileList":
            assert expected == downloaded.dict()[key], f"key = {key}"
    for i in range(len(dblock.dataFileList)):
        for key, expected in dblock.dataFileList[i].dict(exclude_none=True).items():
            assert (
                expected == downloaded.dataFileList[i].dict()[key]
            ), f"i = {i}, key = {key}"


def test_get_dataset(client, derived_dataset, orig_datablock):
    dset = client.get_dataset(derived_dataset.pid)

    assert dset.source_folder == derived_dataset.sourceFolder
    assert dset.creation_time == derived_dataset.creationTime
    assert dset.access_groups == derived_dataset.accessGroups
    assert dset.meta["temperature"] == derived_dataset.scientificMetadata["temperature"]
    assert dset.meta["data_type"] == derived_dataset.scientificMetadata["data_type"]

    for i in range(len(orig_datablock.dataFileList)):
        assert dset.files[i].local_path is None
        assert dset.files[i].size == orig_datablock.dataFileList[i].size
        assert dset.files[i].creation_time == dateutil.parser.parse(
            orig_datablock.dataFileList[i].time
        )


def test_fake_can_disable_functions():
    client = FakeClient(
        disable={
            "get_dataset_model": RuntimeError("custom failure"),
            "get_orig_datablocks": IndexError("custom index error"),
        }
    )
    with pytest.raises(RuntimeError, match="custom failure"):
        client.scicat.get_dataset_model("some-pid")
    with pytest.raises(IndexError, match="custom index error"):
        client.scicat.get_orig_datablocks("some-pid")



import json
from pathlib import Path
import requests
import shutil
import subprocess
import time
from urllib.parse import urljoin

from ..data import load_datasets, load_orig_datablocks


_TEST_BASE = Path(__file__).resolve().parent.parent
_SCICAT_DOCKER_CONFIG = _TEST_BASE / "scicatlive/docker-compose.yaml"
_SCICAT_DATASET_SEED_FILE = Path("seed_db/seed/Dataset.json")
_SCICAT_ORIG_DATABLOCK_SEED_FILE = Path("seed_db/seed/OrigDatablock.json")


# List of required services for tests.
# We only need the backend and API to run tests.
_SERVICES = ("catamel", "mongodb", "mongodb_seed", "reverse-proxy")


def can_connect(scicat_access) -> bool:
    response = requests.post(
        urljoin(scicat_access.url, "Users/login"),
        json=scicat_access.functional_credentials,
    )
    return response.ok


def wait_until_backend_is_live(scicat_access, max_time: float, n_tries: int):
    """
    The containers take a while to be fully live.
    """
    for _ in range(n_tries):
        if can_connect(scicat_access):
            return
        time.sleep(max_time / n_tries)
    if not can_connect(scicat_access):
        raise RuntimeError("Cannot connect to backend")


def _merge_seed_file(target_dir, seed_file_name, custom_seed):
    with open(target_dir / seed_file_name, "r") as f:
        dset_seed = json.load(f)
    dset_seed.extend(custom_seed)
    with open(target_dir / seed_file_name, "w") as f:
        json.dump(dset_seed, f)


def configure(target_dir) -> Path:
    target = Path(target_dir) / "scicatlive"
    shutil.copytree(_SCICAT_DOCKER_CONFIG.parent, target)
    _merge_seed_file(target, _SCICAT_DATASET_SEED_FILE, load_datasets())
    _merge_seed_file(target, _SCICAT_ORIG_DATABLOCK_SEED_FILE, load_orig_datablocks())

    return target / _SCICAT_DOCKER_CONFIG.name


def start_backend_containers(config_file):
    # waiting for scicatlive-mongodb_seed-1 with recreate seems to deadlock
    subprocess.check_call(
        [
            "docker",
            "compose",
            "--file",
            config_file,
            "up",
            "--detach",
            "--force-recreate",
            *_SERVICES,
        ]
    )


def stop_backend_containers(config_file):
    subprocess.check_call(
        ["docker", "compose", "--file", config_file, "down", "--volumes"]
    )



import json
from pathlib import Path

from pyscicat.model import DerivedDataset, OrigDatablock, RawDataset


_BASE_PATH = Path(__file__).resolve().parent


def _process_field(x):
    if isinstance(x, dict) and "$date" in x:
        return x["$date"]
    if isinstance(x, (dict, list)):
        return _process_json(x)
    return x


def _process_json(raw):
    if isinstance(raw, dict):
        processed = {key: _process_field(val) for key, val in raw.items()}
        try:
            processed["pid"] = processed.pop("_id")
        except KeyError:
            pass
    else:
        processed = [_process_field(val) for val in raw]
    return processed


def _load(filename):
    with open(_BASE_PATH / filename, "r") as f:
        return json.load(f)


def load_datasets():
    return _load("datasets.json")


def as_dataset_model(dataset_json):
    ds = _process_json(dataset_json)
    if ds["type"] == "derived":
        return DerivedDataset(**ds)
    return RawDataset(**ds)


def load_orig_datablocks():
    return _load("orig_datablocks.json")


def as_orig_datablock_model(orig_datablock_json):
    processed = _process_json(orig_datablock_json)
    processed["id"] = processed.pop("pid")
    return OrigDatablock(**processed)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen
"""Client to handle communication with SciCat servers."""

from __future__ import annotations
from pathlib import Path
from typing import List, Optional, Union

import pyscicat.client
from pyscicat import model

from .dataset import Dataset
from .typing import FileTransfer


class Client:
    """SciCat client to communicate with a server.

    Clients hold all information needed to communicate with a SciCat instance
    and a filesystem that holds data files (via ``file_transfer``).

    Use :func:`Client.from_token` or :func:`Client.from_credentials` to initialize
    a client instead of the constructor directly.

    See the user guide for typical usage patterns.
    In particular `Downloading Datasets <../../user-guide/downloading.ipynb>`_
    and `Uploading Datasets <../../user-guide/uploading.ipynb>`_.
    """

    def __init__(
        self,
        *,
        client: ScicatClient,
        file_transfer: Optional[FileTransfer],
    ):
        """Initialize a client.

        Do not use directly, instead use :func:`Client.from_token`
        or :func:`Client.from_credentials`!
        """
        self._client = client
        self._file_transfer = file_transfer

    @classmethod
    def from_token(
        cls, *, url: str, token: str, file_transfer: Optional[FileTransfer] = None
    ) -> Client:
        """Create a new client and authenticate with a token.

        Parameters
        ----------
        url:
            URL of the SciCat api.
        token:
            User token to authenticate with SciCat.
        file_transfer:
            Handler for down-/uploads of files.

        Returns
        -------
        :
            A new client.
        """
        return Client(
            client=ScicatClient(pyscicat.client.from_token(base_url=url, token=token)),
            file_transfer=file_transfer,
        )

    @classmethod
    def from_credentials(
        cls,
        *,
        url: str,
        username: str,
        password: str,
        file_transfer: Optional[FileTransfer] = None,
    ) -> Client:
        """Create a new client and authenticate with username and password.

        Parameters
        ----------
        url:
            URL of the SciCat api.
            It should include the suffix `api/vn` where `n` is a number.
        username:
            Name of the user.
        password:
            Password of the user.
        file_transfer:
            Handler for down-/uploads of files.

        Returns
        -------
        :
            A new client.
        """
        return Client(
            client=ScicatClient(
                pyscicat.client.from_credentials(
                    base_url=url, username=username, password=password
                )
            ),
            file_transfer=file_transfer,
        )

    def get_dataset(self, pid: str) -> Dataset:
        """Download a dataset from SciCat.

        Does not download any files.

        Parameters
        ----------
        pid:
            ID of the dataset. Must include the prefix, i.e. have the form
            ``prefix/dataset-id``.

        Returns
        -------
        :
            A new dataset.
        """
        return Dataset.from_models(
            dataset_model=self.scicat.get_dataset_model(pid),
            orig_datablock_models=self.scicat.get_orig_datablocks(pid),
        )

    def upload_new_dataset_now(self, dataset: Dataset) -> Dataset:
        """Upload a dataset as a new entry to SciCat immediately.

        The dataset is inserted as a new entry in the database and will
        never overwrite existing data.
        To this end, the input dataset's ID is ignored and a new one is
        assigned automatically.

        Parameters
        ----------
        dataset:
            The dataset to upload.

        Returns
        -------
        :
            A copy of the input dataset with fields adjusted
            according to the response of the server.

        Raises
        ------
        pyscicat.client.ScicatCommError
            If the upload to SciCat fails.
        RuntimeError
            If the file upload fails or if a critical error is encountered
            and some files or a partial dataset are left on the servers.
            Note the error message if that happens.
        """
        dset = dataset.prepare_as_new()
        dset.pid = dset.pid.without_prefix
        with self.file_transfer.connect_for_upload(dset.pid) as con:
            dset.source_folder = con.source_dir
            for file in dset.files:
                file.source_folder = dset.source_folder
                con.upload_file(local=file.local_path, remote=file.remote_access_path)

            models = dset.make_scicat_models()
            try:
                dataset_id = self.scicat.create_dataset_model(models.dataset)
            except pyscicat.client.ScicatCommError:
                for file in dset.files:
                    con.revert_upload(
                        local=file.local_path, remote=file.remote_access_path
                    )
                raise

        dset.pid = dataset_id
        models.datablock.datasetId = dataset_id
        try:
            self.scicat.create_orig_datablock(models.datablock)
        except pyscicat.client.ScicatCommError as exc:
            raise RuntimeError(
                f"Failed to upload original datablocks for SciCat dataset {dset.pid}:"
                f"\n{exc.args}\nThe dataset and data files were successfully uploaded "
                "but are not linked with each other. Please fix the dataset manually!"
            ) from exc

        return dset

    @property
    def scicat(self) -> ScicatClient:
        """Low level client for SciCat.

        Should typically not be used by users of Scitacean!
        """
        return self._client

    @property
    def file_transfer(self) -> FileTransfer:
        """Stored handler for file down-/uploads."""
        return self._file_transfer

    def download_file(self, *, remote: Union[str, Path], local: Union[str, Path]):
        if self._file_transfer is None:
            raise RuntimeError(
                f"No file transfer handler specified, cannot download file {remote}"
            )
        with self._file_transfer.connect_for_download() as con:
            con.download_file(remote=remote, local=local)

    def upload_file(
        self, *, dataset_id: str, remote: Union[str, Path], local: Union[str, Path]
    ) -> str:
        if self._file_transfer is None:
            raise RuntimeError(
                f"No file transfer handler specified, cannot upload file {local}"
            )
        with self._file_transfer.connect_for_upload(dataset_id) as con:
            return con.upload_file(remote=remote, local=local)


class ScicatClient:
    def __init__(self, client: pyscicat.client.ScicatClient):
        self._client = client

    def get_dataset_model(
        self, pid: str
    ) -> Union[model.DerivedDataset, model.RawDataset]:
        """Fetch a dataset from SciCat.

        Parameters
        ----------
        pid:
            Unique ID of the dataset. Must include the facility ID.

        Returns
        -------
        :
            A model of the dataset. The type is deduced from the received data.

        Raises
        ------
         pyscicat.client.ScicatCommError
            If the dataset does not exist or communication fails for some other reason.
        """
        if dset_json := self._client.datasets_get_one(pid):
            return (
                model.DerivedDataset(**dset_json)
                if dset_json["type"] == "derived"
                else model.RawDataset(**dset_json)
            )
        raise pyscicat.client.ScicatCommError(f"Unable to retrieve dataset {pid}")

    def get_orig_datablocks(self, pid: str) -> List[model.OrigDatablock]:
        """Fetch all orig datablocks from SciCat for a given dataset.

        Parameters
        ----------
        pid:
            Unique ID of the *dataset*. Must include the facility ID.

        Returns
        -------
        :
            Models of the orig datablocks.

        Raises
        ------
         pyscicat.client.ScicatCommError
            If the datablock does not exist or communication
            fails for some other reason.
        """
        if dblock_json := self._client.datasets_origdatablocks_get_one(pid):
            return [_make_orig_datablock(dblock) for dblock in dblock_json]
        raise pyscicat.client.ScicatCommError(
            f"Unable to retrieve orig datablock for dataset {pid}"
        )

    def create_dataset_model(self, dset: model.Dataset) -> str:
        """Create a new dataset in SciCat.

        The dataset PID must be either

        - ``None``, in which case SciCat assigns an ID.
        - An unused id, in which case SciCat uses it for the new dataset.

        If the ID already exists, creation will fail without
        modification to the database.
        Unless the new dataset is identical to the existing one,
        in which case nothing happens.

        Parameters
        ----------
        dset:
            Model of the dataset to create.

        Returns
        -------
        :
            The PID of the new dataset in the catalogue.

        Raises
        ------
         pyscicat.client.ScicatCommError
            If SciCat refuses the dataset or communication
            fails for some other reason.
        """
        return self._client.datasets_create(dset)["pid"]

    def create_orig_datablock(self, dblock: model.OrigDatablock):
        """Create a new orig datablock in SciCat.

        The datablock PID must be either

        - ``None``, in which case SciCat assigns an ID.
        - An unused id, in which case SciCat uses it for the new datablock.

        If the ID already exists, creation will fail without
        modification to the database.

        Parameters
        ----------
        dblock:
            Model of the orig datablock to create.

        Raises
        ------
         pyscicat.client.ScicatCommError
            If SciCat refuses the datablock or communication
            fails for some other reason.
        """
        self._client.datasets_origdatablock_create(dblock)


def _make_orig_datablock(fields):
    files = [model.DataFile(**file_fields) for file_fields in fields["dataFileList"]]
    return model.OrigDatablock(**{**fields, "dataFileList": files})



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen
from contextlib import contextmanager
import os
from pathlib import Path
from typing import Any, Dict, Optional, Union

try:
    from pyfakefs.fake_filesystem import FakeFilesystem
except ImportError:
    FakeFilesystem = Any

from ..pid import PID


class FakeDownloadConnection:
    def __init__(self, fs: Optional[FakeFilesystem], files: Dict[str, bytes]):
        self.files = files
        self.fs = fs

    def download_file(self, *, remote, local):
        if self.fs is not None:
            self.fs.create_file(local, contents=self.files[remote])
        else:
            with open(local, "wb") as f:
                f.write(self.files[remote])


class FakeUploadConnection:
    def __init__(
        self,
        files: Dict[str, bytes],
        reverted: Optional[Dict[str, bytes]],
        dataset_id: PID,
    ):
        self.files = files
        self.reverted = reverted
        self._dataset_id = dataset_id

    @property
    def source_dir(self) -> str:
        return f"/remote/{self._dataset_id.pid}/"

    def _remote_path(self, filename) -> str:
        return os.path.join(self.source_dir, filename)

    def upload_file(self, *, remote: Union[str, Path], local: Union[str, Path]) -> str:
        remote = self._remote_path(remote)
        with open(local, "rb") as f:
            self.files[remote] = f.read()
        return remote

    def revert_upload(self, *, remote: Union[str, Path], local: Union[str, Path] = ""):
        remote = self._remote_path(remote)
        self.reverted[remote] = self.files.pop(remote)


class FakeFileTransfer:
    """Mimic a file down-/upload handler.

    Files are not transferred to a server but loaded into an internal storage.

    It is possible to use a fake file system as implemented by
    `pyfakefs <https://pypi.org/project/pyfakefs/>`_.

    Examples
    --------
    Using the ``fs`` fixture from pyfakefs in pytest:

    .. code-block:: python

        def test_upload(fs):
            client = FakeClient.from_token(
                url="...",
                token="...",
                file_transfer=FakeFileTransfer(fs=fs))
            dset = ...
            client.upload_new_dataset_now(dset)
            assert client.file_transfer.files[expected_remote_path] == file_content

    See Also
    --------
    scitacean.testing.client.FakeClient:
        Client to mimic a SciCat server.
    """

    def __init__(
        self,
        *,
        fs: Optional[FakeFilesystem] = None,
        files: Optional[Dict[str, bytes]] = None,
        reverted: Optional[Dict[str, bytes]] = None,
    ):
        """Initialize a file transfer.

        Parameters
        ----------
        fs:
            Fake filesystem. If given, files are down-/uploaded to/from this
            filesystem instead of the real one.
            If set to ``None``, the real filesystem is used.
        files:
            Initial files stored "on the server".
            Maps file names to contents.
        reverted:
            Files that have been uploaded and subsequently been removed.
        """
        self.fs = fs
        self.files = {} if files is None else files
        self.reverted = {} if reverted is None else reverted

    @contextmanager
    def connect_for_download(self):
        yield FakeDownloadConnection(fs=self.fs, files=self.files)

    @contextmanager
    def connect_for_upload(self, dataset_id: PID):
        yield FakeUploadConnection(
            files=self.files, reverted=self.reverted, dataset_id=dataset_id
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen

import logging


def get_logger():
    return logging.getLogger("scitacean")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

from .client import Client
from .dataset import Dataset
from .file import File

__all__ = ("Client", "Dataset", "File")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen
"""Main dataset structure."""

from __future__ import annotations

import dataclasses
from collections import namedtuple
import html
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
import typing
from uuid import uuid4

from pyscicat.model import DerivedDataset, OrigDatablock, RawDataset

from .file import File
from .pid import PID
from ._dataset_fields import DatasetFields

SciCatModels = namedtuple("SciCatModels", ["dataset", "datablock"])


# TODO handle orig vs non-orig datablocks
# TODO add derive method
class Dataset(DatasetFields):
    """Store metadata and links to files.

    ``Dataset`` has a large number of attributes that map onto the fields
    in a SciCat dataset.
    See also :class:`scitacean._dataset_fields.DatasetFields` for a complete list.
    For the time being, see also :class:`pyscicat.model.DerivedDataset` and
    :class:`pyscicat.model.RawDataset` for which fields are required by which
    dataset type.
    """

    # If self._datablock is None, the dataset does not exist on remote
    def __init__(
        self,
        *,
        files: List[File],
        datablock: Optional[OrigDatablock],
        meta: Dict[str, Any],
        pid: Optional[Union[PID, str]],
        **kwargs,
    ):
        """Use :meth:`Dataset.new` instead."""
        super().__init__(**kwargs)
        self._files = files
        self._datablock = datablock
        self._meta = dict(meta)
        self.pid = pid

    @classmethod
    def new(
        cls,
        *,
        model: Optional[Union[DerivedDataset, RawDataset]] = None,
        meta: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Dataset:
        """Create a new Dataset.

        The dataset has no files and is not linked or uploaded to SciCat.

        Parameters
        ----------
        model:
            pydantic model of the dataset.
            Contents are incorporated into the new Dataset.
        meta:
            dict of scientific metadata.
        kwargs:
            Merged with ``model`` to fill the dataset fields.
            Items in ``kwargs`` override items of the same key in ``model``.

        Returns
        -------
        :
            A new dataset.
        """
        # TODO some model fields are ignores, e.g. size
        model_dict = cls._map_model_to_field_dict(model) if model is not None else {}
        model_dict.update(kwargs)
        model_dict["pid"] = model.pid if model is not None else None
        meta = {} if meta is None else dict(meta)
        if model is not None and model.scientificMetadata:
            meta.update(model.scientificMetadata)
            model.scientificMetadata = None
        return Dataset(files=[], datablock=None, meta=meta, **model_dict)

    @classmethod
    def from_models(
        cls,
        *,
        dataset_model: Union[DerivedDataset, RawDataset],
        orig_datablock_models: List[OrigDatablock],
    ):
        """Create a new dataset from fully filled in models.

        Parameters
        ----------
        dataset_model:
            Fields, including scientific metadata are filled from this model.
        orig_datablock_models:
            File links are populated from this model.

        Returns
        -------
        :
            A new dataset.
        """
        if len(orig_datablock_models) != 1:
            raise NotImplementedError(
                f"Got {len(orig_datablock_models)} original datablocks for "
                f"dataset {dataset_model.pid} but only support for one is implemented."
            )
        dblock = orig_datablock_models[0]

        meta = dataset_model.scientificMetadata
        if meta is None:
            meta = {}
        return Dataset(
            files=[
                File.from_scicat(file, source_folder=dataset_model.sourceFolder)
                for file in dblock.dataFileList
            ],
            datablock=dblock,
            meta=meta,
            pid=dataset_model.pid,
            **{
                k: v
                for k, v in cls._map_model_to_field_dict(dataset_model).items()
                if k != "scientificMetadata"
            },
        )

    @property
    def meta(self) -> Dict[str, Any]:
        """Dictionary of scientific metadata."""
        return self._meta

    @property
    def pid(self) -> Optional[PID]:
        """ID of the dataset."""
        return self._pid

    @pid.setter
    def pid(self, pid: Optional[Union[PID, str]]):
        """Set the ID of the dataset."""
        self._pid = pid if isinstance(pid, PID) or pid is None else PID.parse(pid)

    @property
    def size(self) -> int:
        """Combined size of all files."""
        return sum(file.size for file in self.files)

    @property
    def files(self) -> Tuple[File, ...]:
        """Currently linked files."""
        return tuple(self._files)

    def add_files(self, *files: File):
        """Add files to the dataset."""
        self._files.extend(files)

    def add_local_files(
        self,
        *paths: Union[str, Path],
        base_path: Union[str, Path] = "",
    ):
        """Add files on the local file system to the dataset.

        Parameters
        ----------
        paths:
            Local paths to the files.
        base_path:
            The remote paths will be set up according to
            ``remote = [path.relative_to(base_path) for path in paths]``.
        """
        self.add_files(*(File.from_local(path, base_path=base_path) for path in paths))

    def make_scicat_models(self) -> SciCatModels:
        """Build models to send to SciCat.

        Creates both a model for that dataset and orig datablocks.

        Returns
        -------
        :
            Created models.
        """
        if self.pid is None:
            raise ValueError(
                "The dataset PID must be set before creating SciCat models."
            )
        total_size = sum(file.model.size for file in self.files)
        datablock = OrigDatablock(
            size=total_size,
            dataFileList=[file.model for file in self.files],
            datasetId=str(self.pid),
            ownerGroup=self.owner_group,
            accessGroups=self.access_groups,
        )
        try:
            model = self._map_to_model(
                number_of_files=len(self.files) or None,
                number_of_files_archived=None,
                packed_size=None,
                pid=str(self.pid) if self.pid is not None else None,
                size=total_size or None,
                scientific_metadata=self.meta or None,
            )
        except KeyError as err:
            raise TypeError(
                f"Field {err} is not allowed in {self.dataset_type} datasets"
            ) from None
        return SciCatModels(dataset=model, datablock=datablock)

    def prepare_as_new(self):
        """Return a copy of this dataset that looks like a new dataset."""
        dset = Dataset(
            files=self._files,
            datablock=self._datablock,
            meta=self._meta,
            pid=PID(pid=str(uuid4()), prefix=None),
            **dataclasses.asdict(self),
        )
        dset.history = None
        dset.updated_at = None
        dset.updated_by = None
        return dset

    def _repr_html_(self):
        rows = "\n".join(
            _format_field(self, field) for field in dataclasses.fields(self)
        )
        return f"""<table>
        <tr>
            <th>Name</th><th>Required</th><th>Type</th><th>Value</th><th>Description</th>
        </tr>
        {rows}
    </table>"""


def _format_field(dset, field) -> str:
    name = field.name
    required = "*" if not _is_optional(field.type) else ""
    value = html.escape(str(getattr(dset, field.name)))
    typ = _format_type(field.type)
    description = html.escape("N/A")
    return (
        "<tr><td>"
        + "</td><td>".join((name, required, typ, value, description))
        + "</tr></td>"
    )


def _format_type(typ) -> str:
    if _is_optional(typ):
        typ = next(iter(arg for arg in typing.get_args(typ) if not _is_none_type(arg)))

    if typ == str:
        return "str"
    if typ == bool:
        return "bool"
    if typ == List[str]:
        return "list[str]"
    return html.escape(str(typ))


def _is_none_type(x) -> bool:
    # We want to actually compare types but flake8 doesn't like it.
    return x == type(None)  # noqa: [E721]


def _is_optional(typ):
    return typing.get_origin(typ) == Union and type(None) in typing.get_args(typ)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# @author Jan-Lukas Wynen

from pathlib import Path
from typing import ContextManager, Protocol, Union

from .pid import PID


class DownloadConnection(Protocol):
    """An open connection to the file server for downloads."""

    def download_file(self, *, remote: Union[str, Path], local: Union[str, Path]):
        """Download a file from the file server.

        Parameters
        ----------
        remote:
            The full path to the file on the server.
        local:
            Desired path of the file on the local filesystem.
        """


class Downloader(Protocol):
    """Handler for file downloads."""

    def connect_for_download(self) -> ContextManager[DownloadConnection]:
        """Open a connection to the file server.

        Returns
        -------
        :
            A connection object that can download files.
        """


class UploadConnection(Protocol):
    """An open connection to the file server for uploads."""

    source_dir: str
    """Files are uploaded to this directory / location."""

    def upload_file(self, *, remote: Union[str, Path], local: Union[str, Path]) -> str:
        """Upload a file to the file server.

        Parameters
        ----------
        remote:
            The file needs to be uploaded to ``source_dir/remote``.
        local:
            Path of the file on the local filesystem.

        Returns
        -------
        :
            The full remote path ``source_dir/remote``.
        """

    def revert_upload(self, *, remote: Union[str, Path], local: Union[str, Path] = ""):
        """Delete a file uploaded by upload_file.

        Only files uploaded by the same connection object may be handled.

        Parameters
        ----------
        remote:
            The file needs to be uploaded to ``source_dir/remote``.
        local:
            Path of the file on the local filesystem.
        """


class Uploader(Protocol):
    """Handler for file uploads."""

    def connect_for_upload(self, dataset_id: PID) -> ContextManager[UploadConnection]:
        """Open a connection to the file server.

        Parameters
        ----------
        dataset_id:
            ID of the dataset whose file will be uploaded.

        Returns
        -------
        :
            A connection object that can upload files.
        """


class FileTransfer(Downloader, Uploader):
    """Handler for file down-/uploads."""
